"""Contains the base Layer class, from which all layers inherit.
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import numpy as np


class Pooling():
    @staticmethod
    def init(layer, input_shape):
        pass
    @staticmethod
    def forword(layer, flow_data):
        # Z = np.dot(layer['weight'], flow_data) + layer['bias']
        return flow_data
    @staticmethod
    def backword(flow_data, layer, config):
        return flow_data
        # layer["weight_gradient"] = (1 / config['batch_size']) * np.dot(flow_data, layer['input'].T)
        # layer["bias_gradient"] = (1 / config['batch_size']) * np.sum(flow_data, axis=1, keepdims=True)
        # flow_data = np.dot(layer['weight'].T, flow_data)
        # return layer, flow_data
